/*
PRICE was written by Graham Ruby.
This file is part of PRICE.

PRICE is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

PRICE is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with PRICE.  If not, see <http://www.gnu.org/licenses/>.
*/

/*
This is an interface for a nucleic acid sequence whose nucleotide identities
are scored.  Scores are representative of the probability of a base having
been mis-called, and are the negative base-2 log of that probabilty.  Any
non-A,T,C,G nucleotide must have a score of 0 as it has not been properly
called and therefore the probability that it is a mis-call is 1.

SPEC FIELDS:
seq: the sequence in nucleic acids (A,T,C,G, or N)
scoreList: the scores of the nucleotides (negative ints)
 */

#ifndef SCOREDSEQ_H
#define SCOREDSEQ_H
# include <vector>
# include <string>
using std::string;
using std::vector;


class ScoredSeq {

 public:

  static ScoredSeq* getScoredSeq(char* seq, float score, long size);
  static ScoredSeq* getScoredSeq(char* seq, float* scores, long size);
  static ScoredSeq* getScoredSeq(char* seq, float* scores, float* links, long size);
  static ScoredSeq* copyShallowSeq(ScoredSeq* seq, char sense);
  virtual ~ScoredSeq();

  /* returns the nucleotide identity at the indicated position
     on the indicated strand.  The position is zero-indexed from the
     5p end of the sequence; i.e. for the nucleotide at position n
     on the '+' strand, its reverse complement would be obtained by
     getting the nucleotide at position (lengthOfSeq - n - 1) from the
     '-' strand.  Remember: numbers always ascend in the 5p->3p direction.
     RETURNS: nucleotide identity
     REQUIRES: sense is '+' or '-'
     THROWS: ??? if position >= length of this
     PARAMS: position: zero-indexed, nums ascend 5p->3p
             sense: '+' or '-'
  */
  virtual char nucAtPosition(long position, char sense) = 0;
  // optimized to not have to pass the sense char and not have to ask which sense it is
  virtual char nucAtPosPlus(long position) = 0;
  virtual char nucAtPosMinus(long position) = 0;

  /* returns the score for the nucleotide called at the
     indicated position on the indicated strand, numbered 5p->3p
     on that strand
     RETURNS: nucleotide confidence score
     REQUIRES: sense is '+' or '-'
     THROWS: ??? if position >= length of this
     PARAMS: position: zero-indexed, nums ascend 5p->3p
             sense: '+' or '-'
  */
  virtual float scoreAtPosition(long position, char sense) = 0;
  virtual float scoreAtPosPlus(long position) = 0;
  virtual float scoreAtPosMinus(long position) = 0;

  virtual float linkAfterPosition(long position, char sense) = 0; // must be <= nucleotide counts
  virtual float linkAfterPosPlus(long position) = 0; // must be <= nucleotide counts
  virtual float linkAfterPosMinus(long position) = 0; // must be <= nucleotide counts

  /* gets the full nucleotide sequence as a string, presented 5p->3p
     RETURNS: the nucelotide DNA sequence (composed of A,T,C,G,N)
     REQUIRES: sense is '+' or '-'
     PARAMS: sense: '+' or '-'
  */
  virtual char* getSeq(char sense) = 0;
  virtual float* getScores(char sense) = 0;
  virtual float* getLinks(char sense) = 0;

  virtual char* getSubseq(long position, long length, char sense) = 0;
  virtual float* getSubScores(long position, long length, char sense) = 0;
  virtual float* getSubLinks(long position, long length, char sense) = 0;

  /* has no effects on state; purely an optimization tool that can make
     retrieval of the sequence or scores using the above methods more efficient
     later in some implementations.
     EFFECTS: none
  */
  virtual void buffer() = 0;
  /* for a multi-nested seq, allows just the bottom level
   * to be buffered */
  virtual void bottomBuffer() = 0;

  /* has no effects on state; purely an optimization tool that can save
     memory in some implementations if called after sequence/score info
     is accessed.
     EFFECTS: none
  */
  virtual void unbuffer() = 0;

  /* RETURNS: length of the sequence in number of nucleotides */
  virtual long size() = 0;

  /* Identifies whether or not the sequence has a paired end.  Paired-ends
     are defined as sequences that derive from the same DNA template as one 
     another and derive from opposing strands.  they may overlap, but the 
     non-sequencing-primer-derived sequences will always be mutually downstream
     of one another (at worst, perfectly overlapping). */
  virtual bool hasPairedEnd() = 0;

  /* if there is a paired end, it can be provided as a ScoredSeq by this method.
     See hasPairedEnd() comments for a description of paired ends.
     THROWS: ??? if this does not have a paired-end.
  */
  virtual ScoredSeq * getPairedEnd() = 0;
  virtual ScoredSeq * getTempPairedEnd() = 0;

  /* returns a copy of the ScoredSeq with shallow scope and so can be destroyed without adverse effects */
  virtual ScoredSeq * shallowCopy() = 0;

  /* returns a copy of the ScoredSeq with shallow scope and also with opposite sense as the original */
  virtual ScoredSeq * flipCopy() = 0;

  /* for wrapper classes */
  virtual bool isNested() = 0;

  /* these methods are supported in all cases even if they are only meaningfully different from their
   * shallow counterparts (destructor and unbuffer) for nested classes */
  virtual void deepDelete() = 0;
  virtual void deepUnbuffer() = 0;

  static char* reverseComplement(char* seq, long length);

};

#endif
